# coding=utf-8
import git
from os.path import expanduser
from mock import patch, MagicMock

from jig.tests.testcase import CommandTestCase, result_with_hint
from jig.commands import sticky
from jig.exc import (
    ForcedExit, JigUserDirectoryError, GitConfigError,
    InitTemplateDirAlreadySet, GitTemplatesMissing, GitHomeTemplatesExists)
from jig.commands.hints import (
    INIT_TEMPLATE_DIR_ALREADY_SET, GIT_TEMPLATES_MISSING,
    GIT_HOME_TEMPLATES_EXISTS)


class TestStickyCommand(CommandTestCase):

    """
    Test the sticky command.

    """
    command = sticky.Command

    def setUp(self):
        super(TestStickyCommand, self).setUp()

        self.mocks = {
            'create_auto_init_templates': MagicMock(),
            'set_templates_directory': MagicMock()
        }

        self._patches = []

    def _start_patches(self):
        assert len(self._patches) == 0

        for function, mock_function in self.mocks.items():
            patched = patch(
                'jig.commands.sticky.{0}'.format(function),
                new=mock_function
            )

            patched.start()
            self._patches.append(patched)

    def run_command(self, *args, **kwargs):
        """
        Make sure that our patches have started before we run a command.
        """
        self._start_patches()

        return super(TestStickyCommand, self).run_command(*args, **kwargs)

    def tearDown(self):
        for patches in self._patches:
            patches.stop()

    def test_command_succeeds(self):
        """
        Successful command returns a message that informs the user.
        """
        self.run_command()

        self.assertResults(
            u'Jig has been setup to run everytime you clone.',
            self.output)

    def test_fails_create_auto_init_templates(self):
        """
        A failure to auto-init is formatted correctly.
        """
        self.mocks['create_auto_init_templates'].side_effect = \
            JigUserDirectoryError('Error')

        with self.assertRaises(ForcedExit):
            self.run_command()

        self.assertResults(
            u'Error',
            self.error)

    def test_templates_missing(self):
        """
        No Git templates can be found.
        """
        self.mocks['create_auto_init_templates'].side_effect = \
            GitTemplatesMissing()

        with self.assertRaises(ForcedExit):
            self.run_command()

        self.assertResults(
            result_with_hint(
                u'Unable to find templates.',
                GIT_TEMPLATES_MISSING),
            self.error)

    def test_home_templates_exist(self):
        """
        A templates directory already exists in ~/.jig/git
        """
        self.mocks['create_auto_init_templates'].side_effect = \
            GitHomeTemplatesExists('~/.jig/git/templates')

        with self.assertRaises(ForcedExit):
            self.run_command()

        self.assertResults(
            result_with_hint(
                u'~/.jig/git/templates already exists',
                GIT_HOME_TEMPLATES_EXISTS),
            self.error)

    def test_init_templatesdir_already_set(self):
        """
        Git is already configured with a init.templatedir
        """
        self.mocks['set_templates_directory'].side_effect = \
            InitTemplateDirAlreadySet('/tmp/templates')

        with self.assertRaises(ForcedExit):
            self.run_command()

        self.assertResults(
            result_with_hint(
                u'Git configuration for init.templatedir is /tmp/templates',
                INIT_TEMPLATE_DIR_ALREADY_SET),
            self.error)

    def test_git_config_error(self):
        """
        A failure to read or write to the Git config.
        """
        self.mocks['set_templates_directory'].side_effect = \
            GitConfigError(git.exc.GitCommandError(
                'git config', 1, 'error'))

        with self.assertRaises(ForcedExit):
            self.run_command()

        self.assertResults(
            u'Problem when running git config: error',
            self.error)



# coding=utf-8
from tempfile import mkdtemp
from contextlib import nested

from mock import patch

from jig.tests.testcase import (
    CommandTestCase, PluginTestCase, result_with_hint)
from jig.plugins import set_jigconfig
from jig.exc import ForcedExit
from jig.output import ATTENTION
from jig.commands import runnow
from jig.commands.hints import GIT_REPO_NOT_INITIALIZED


class TestRunNowCommand(CommandTestCase, PluginTestCase):

    """
    Test the runnow command.

    """
    command = runnow.Command

    def test_no_changes(self):
        """
        No changes have been made to the Git repository.
        """
        self._add_plugin(self.jigconfig, 'plugin01')
        set_jigconfig(self.gitrepodir, config=self.jigconfig)

        # Create the first commit
        self.commit(self.gitrepodir, 'a.txt', 'a')

        with self.assertRaises(SystemExit) as ec:
            self.run_command(self.gitrepodir)

        self.assertSystemExitCode(ec.exception, 0)

        self.assertEqual(
            u'No staged changes in the repository, '
            u'skipping jig.\n', self.output)

    def test_changes(self):
        """
        Changes are made and the plugin runs and gives us output.
        """
        self._add_plugin(self.jigconfig, 'plugin01')
        set_jigconfig(self.gitrepodir, config=self.jigconfig)

        # Create staged changes
        self.commit(self.gitrepodir, 'a.txt', 'a')
        self.stage(self.gitrepodir, 'b.txt', 'b')

        with nested(
            patch('jig.runner.sys'),
            self.assertRaises(SystemExit)
        ) as (r_sys, ec):
            # Raise the error to halt execution like the real sys.exit would
            r_sys.exit.side_effect = SystemExit

            self.run_command(self.gitrepodir)

        r_sys.exit.assert_called_once_with(0)

        self.assertResults(u"""
            ▾  plugin01

            ⚠  line 1: b.txt
                b is +

            {0}  Jig ran 1 plugin
                Info 0 Warn 1 Stop 0
            """.format(ATTENTION), self.output)

    def test_specific_plugin_installed(self):
        """
        A specific plugin can be ran if it's installed.
        """
        self._add_plugin(self.jigconfig, 'plugin01')
        set_jigconfig(self.gitrepodir, config=self.jigconfig)

        # Create staged
        self.commit(self.gitrepodir, 'a.txt', 'a')
        self.stage(self.gitrepodir, 'b.txt', 'b')

        with nested(
            patch('jig.runner.sys'),
            self.assertRaises(SystemExit)
        ) as (r_sys, ec):
            # Raise the error to halt execution like the real sys.exit would
            r_sys.exit.side_effect = SystemExit

            self.run_command('--plugin plugin01 {0}'.format(self.gitrepodir))

        self.assertResults(u"""
            ▾  plugin01

            ⚠  line 1: b.txt
                b is +

            {0}  Jig ran 1 plugin
                Info 0 Warn 1 Stop 0
            """.format(ATTENTION), self.output)

    def test_specific_plugin_not_installed(self):
        """
        A specific plugin can be ran but it's not installed.
        """
        self._add_plugin(self.jigconfig, 'plugin01')
        set_jigconfig(self.gitrepodir, config=self.jigconfig)

        # Create staged
        self.commit(self.gitrepodir, 'a.txt', 'a')
        self.stage(self.gitrepodir, 'b.txt', 'b')

        with nested(
            patch('jig.runner.sys'),
            self.assertRaises(SystemExit)
        ) as (r_sys, ec):
            # Raise the error to halt execution like the real sys.exit would
            r_sys.exit.side_effect = SystemExit

            self.run_command(
                '--plugin notinstalled {0}'.format(self.gitrepodir))

        # A plugin which is not installed was requested so not output
        self.assertEqual('', self.output)

    def test_handles_error(self):
        """
        An un-initialized jig Git repository provides an error message.
        """
        with self.assertRaises(ForcedExit):
            self.run_command(mkdtemp())

        self.assertResults(
            result_with_hint(
                u'This repository has not been initialized.',
                GIT_REPO_NOT_INITIALIZED),
            self.error)



import git
from git.exc import GitCommandError, BadObject

from jig.exc import GitRevListFormatError, GitRevListMissing


def parse_rev_range(repository, rev_range):
    """
    Convert revision range to two :class:`git.objects.commit.Commit` objects.

    :param string repository: path to the Git repository
    :param string rev_range: Double dot-separated revision range, like
        "FOO..BAR"
    :returns: the two commits representing the range
    :rtype: tuple
    """
    rev_pair = rev_range.split('..')

    if len(rev_pair) != 2 or not all(rev_pair):
        raise GitRevListFormatError(rev_range)

    rev_a, rev_b = rev_pair

    try:
        repo = git.Repo(repository)

        commit_a = repo.commit(rev_a)
        commit_b = repo.commit(rev_b)

        return commit_a, commit_b
    except (BadObject, GitCommandError):
        raise GitRevListMissing(rev_range)



from os import symlink
from os.path import join, realpath
from functools import wraps
from textwrap import dedent
from pprint import PrettyPrinter
from operator import itemgetter

from mock import Mock
from git import Repo

from jig.tests.testcase import JigTestCase
from jig.diffconvert import describe_diff, DiffType, GitDiffIndex
from jig.tools import cwd_bounce


def assertDiff(func):
    """
    Decorator used to test diffs.

    Uses ``yield`` in the following way:

        @assertDiff
        def test_my_diff(self):
            yield 'one'

            yield 'two'

            yield [(1, '-', 'one'), (1, '+', 'two')]

    The order of the yields are:

        1. Original value (a)
        2. After it's edited (b)
        3. The expected difference
    """
    pp = PrettyPrinter().pformat

    @wraps(func)
    def wrapper(self, **kwargs):
        queue = func(self, **kwargs)

        a = queue.next()
        b = queue.next()
        expected = queue.next()

        a = dedent(a).strip()
        b = dedent(b).strip()

        actual = [i for i in describe_diff(a, b)]
        if not expected == actual:   # pragma: no cover
            self.fail('Diff does not match:\nexpected\n{}\nactual\n{}'.format(
                pp(expected),
                pp(actual)))

    return wrapper


class TestDescribeDiff(JigTestCase):

    """
    Test our diff description method.

    """
    @assertDiff
    def test_all_addition(self):
        """
        All lines are being added.
        """
        yield ''

        yield '''
            one
            two
            three'''

        yield [
            (1, '+', 'one'),
            (2, '+', 'two'),
            (3, '+', 'three')]

    @assertDiff
    def test_add_blank_lines(self):
        """
        Lines added are just blank lines.
        """
        yield '''
            one
            two
            three'''

        yield '''
            one


            two
            three'''

        # This is a bit counter-intuitive, but correct
        yield [
            (1, ' ', 'one'),
            (2, '+', ''),
            (3, '+', ''),
            (4, ' ', 'two'),
            (5, ' ', 'three')]

    @assertDiff
    def test_all_same(self):
        """
        No changes.
        """
        yield '''
            one
            two
            three'''

        yield '''
            one
            two
            three'''

        yield [
            (1, ' ', 'one'),
            (2, ' ', 'two'),
            (3, ' ', 'three')]

    @assertDiff
    def test_one_insert(self):
        """
        Just one line inserted.
        """
        yield '''
            one
            two
            three'''

        yield '''
            one
            two
            2.5
            three'''

        yield [
            (1, ' ', 'one'),
            (2, ' ', 'two'),
            (3, '+', '2.5'),
            (4, ' ', 'three')]

    @assertDiff
    def test_one_delete(self):
        """
        Just one deleted.
        """
        yield '''
            one
            two
            three
            four'''

        yield '''
            one
            two
            four'''

        yield [
            (1, ' ', 'one'),
            (2, ' ', 'two'),
            (3, '-', 'three'),
            (3, ' ', 'four')]

    @assertDiff
    def test_one_insert_delete(self):
        """
        One insert, one delete.
        """
        yield '''
            one
            two
            three
            four'''

        yield '''
            one
            two
            3
            four'''

        yield [
            (1, ' ', 'one'),
            (2, ' ', 'two'),
            (3, '-', 'three'),
            (3, '+', '3'),
            (4, ' ', 'four')]

    @assertDiff
    def test_one_character_change(self):
        """
        A single character changed.
        """
        yield '''
            one
            two
            three
            four'''

        yield '''
            one
            two
            thr3e
            four'''

        yield [
            (1, ' ', 'one'),
            (2, ' ', 'two'),
            (3, '-', 'three'),
            (3, '+', 'thr3e'),
            (4, ' ', 'four')]

    @assertDiff
    def test_complex_01(self):
        """
        Complex example with several changes.
        """
        yield '''
            one
            two
            three
            three-and-a-smidge
            four'''

        yield '''
            one
            1.5
            two
            three

            four'''

        yield [
            (1, ' ', 'one'),
            (2, '+', '1.5'),
            (3, ' ', 'two'),
            (4, ' ', 'three'),
            (4, '-', 'three-and-a-smidge'),
            (5, '+', ''),
            (6, ' ', 'four')]


class TestDiffType(JigTestCase):

    """
    Detect diff type from :py:class:`Git.Diff` objects.

    """
    def test_add(self):
        """
        Add type.
        """
        diff = Mock()
        diff.new_file = True

        self.assertEqual(DiffType.A, DiffType.for_diff(diff))

    def test_deleted(self):
        """
        Deleted type.
        """
        diff = Mock()
        diff.new_file = False
        diff.deleted_file = True

        self.assertEqual(DiffType.D, DiffType.for_diff(diff))

    def test_renamed(self):
        """
        Renamed type.
        """
        diff = Mock()
        diff.new_file = False
        diff.deleted_file = False
        diff.renamed = True

        self.assertEqual(DiffType.R, DiffType.for_diff(diff))

    def test_modified(self):
        """
        Modified type.
        """
        diff = Mock()
        diff.new_file = False
        diff.deleted_file = False
        diff.renamed = False
        diff.a_blob = 'blob a'
        diff.b_blob = 'blob b'

        self.assertEqual(DiffType.M, DiffType.for_diff(diff))

    def test_unknown(self):
        """
        Unknown type.
        """
        diff = Mock()
        diff.new_file = False
        diff.deleted_file = False
        diff.renamed = False
        diff.a_blob = False
        diff.b_blob = False

        self.assertEqual(DiffType.U, DiffType.for_diff(diff))


class TestGitDiffIndex(JigTestCase):

    """
    Test converting Git changes to JSON.

    """
    def setUp(self):
        super(TestGitDiffIndex, self).setUp()

        repo, working_dir, diffs = self.repo_from_fixture('repo01')

        self.testrepo = repo
        self.testrepodir = working_dir
        self.testdiffs = diffs

    def test_new_file(self):
        """
        Handles new files.
        """
        gdi = self.git_diff_index(self.testrepo, self.testdiffs[0])

        self.assertEqual(1, len(list(gdi.files())))

        file1 = gdi.files().next()

        # This one is relative to the Git repo
        self.assertEqual('argument.txt', file1['name'])
        # It should be added because this is a new file
        self.assertEqual('added', file1['type'])
        # This one is the full path to the file
        self.assertEqual(
            realpath(join(self.testrepodir, 'argument.txt')),
            realpath(file1['filename']))

    def test_modified(self):
        """
        Handles modified files.
        """
        gdi = self.git_diff_index(self.testrepo, self.testdiffs[1])

        self.assertEqual(1, len(list(gdi.files())))

        file1 = gdi.files().next()
        diff = [i for i in file1['diff']]
        difftypes = set([i[1] for i in diff])

        # File was changed
        self.assertEqual('modified', file1['type'])

        # We should have every kind of modification
        # Same lines, additions, and subtractions
        self.assertEqual(
            set([' ', '+', '-']),
            difftypes)

        # And we have a list of differences as expected
        self.assertEqual(47, len(diff))

    def test_deleted_file(self):
        """
        Handles deleted files.
        """
        gdi = self.git_diff_index(self.testrepo, self.testdiffs[2])

        self.assertEqual(1, len(list(gdi.files())))

        file1 = gdi.files().next()
        diff = [i for i in file1['diff']]
        difftypes = set([i[1] for i in diff])

        # File was deleted
        self.assertEqual('deleted', file1['type'])

        # Each line should be a removal
        self.assertEqual(
            set(['-']),
            difftypes)

        self.assertEqual(35, len(diff))

    def test_multiple_changes(self):
        """
        Handles multiple files changed.
        """
        gdi = self.git_diff_index(self.testrepo, self.testdiffs[3])

        self.assertEqual(2, len(list(gdi.files())))

        files = sorted(
            [i for i in gdi.files()],
            key=itemgetter('name'))

        self.assertEqual(
            'famous-deaths.txt',
            files[0]['name'])

        self.assertEqual(
            'italian-lesson.txt',
            files[1]['name'])

    def test_name_contains_subdirectories(self):
        """
        If sub-directories are involved, those are included properly.
        """
        gdi = self.git_diff_index(self.testrepo, self.testdiffs[4])

        # Since we've moved the file Git will see this as a deletion of 2 files
        # plus the addition of 2 files, so it makes our count 4.
        self.assertEqual(4, len(list(gdi.files())))

        files = sorted(
            [i for i in gdi.files()],
            key=itemgetter('name'))

        # Make sure that the name contains our sub-directory.
        self.assertEqual(
            'scripts/famous-deaths.txt',
            files[2]['name'])

        self.assertEqual(
            'scripts/italian-lesson.txt',
            files[3]['name'])

    def test_binary_diff(self):
        """
        Binary files are ignored.
        """
        gdi = self.git_diff_index(self.testrepo, self.testdiffs[5])

        # We should see our file
        self.assertEqual(1, len(list(gdi.files())))

        # But we don't include the diff since it's binary data
        self.assertEqual([], gdi.files().next()['diff'])

    def test_ignores_jig_directory(self):
        """
        Does not include anything in the .jig directory.
        """
        gdi = self.git_diff_index(self.testrepo, self.testdiffs[6])

        # We should see our file
        self.assertEqual(0, len(list(gdi.files())))

    def test_symlinks(self):
        """
        Symlinks are ignored because they are not real files.
        """
        self.commit(self.gitrepodir, 'text/a.txt', 'a')
        self.commit(self.gitrepodir, 'text/b.txt', 'b')
        self.commit(self.gitrepodir, 'text/c.txt', 'c')

        # Create the symlink that should be ignored by GitDiffIndex
        with cwd_bounce(self.gitrepodir):
            symlink('text', 'also_text')

        # We have to do this without our testcase since it's a special
        # situation.
        repo = Repo(self.gitrepodir)
        repo.git.add('also_text')

        # The symlink is staged, time to convert the diff
        gdi = GitDiffIndex(self.gitrepodir, repo.head.commit.diff())

        # If we ignored the symlink, which we should, there should be no files
        self.assertEqual(0, len(list(gdi.files())))



from os.path import join
from ConfigParser import SafeConfigParser
from collections import namedtuple
from textwrap import TextWrapper

from jig.conf import PLUGIN_CONFIG_FILENAME
from jig.exc import CommandError, ConfigKeyInvalid
from jig.commands.base import BaseCommand, plugins_by_bundle
from jig.commands.hints import NO_PLUGINS_INSTALLED, CHANGE_PLUGIN_SETTINGS
from jig.plugins import (
    get_jigconfig, set_jigconfig, PluginManager)

try:
    import argparse
except ImportError:   # pragma: no cover
    from backports import argparse

_parser = argparse.ArgumentParser(
    description='Manage settings for installed Jig plugins',
    usage='jig config [-h] ACTION')

_subparsers = _parser.add_subparsers(
    title='actions',
    description='available commands to manage settings')

_listparser = _subparsers.add_parser(
    'list', help='list all settings',
    usage='jig config list [-h] [-r GITREPO]')
_listparser.add_argument(
    '--gitrepo', '-r', default='.', dest='path',
    help='Path to the Git repository, default current directory')
_listparser.set_defaults(subcommand='list')

_aboutparser = _subparsers.add_parser(
    'about', help='learn about a plugin\'s settings',
    usage='jig config about [-h] [-r GITREPO]')
_aboutparser.add_argument(
    '--gitrepo', '-r', default='.', dest='path',
    help='Path to the Git repository, default current directory')
_aboutparser.set_defaults(subcommand='about')

_setparser = _subparsers.add_parser(
    'set', help='set a single setting for an installed plugin',
    usage='jig config set [-h] [-r GITREPO] KEY VALUE')
_setparser.add_argument(
    '--gitrepo', '-r', default='.', dest='path',
    help='Path to the Git repository, default current directory')
_setparser.add_argument(
    'key',
    help='Setting key which is a dot-separated string of the '
    'bundle name, plugin name, and setting name')
_setparser.add_argument(
    'value',
    help='Value for the specified settings')
_setparser.set_defaults(subcommand='set')


try:
    from collections import OrderedDict
except ImportError:   # pragma: no cover
    from ordereddict import OrderedDict


def _get_plugin_config_section(plugin_dir, section):
    """
    Get a section of a plugin's config.

    :param string plugin_dir: the directory where the plugin's is found
    :param string section: name of the section to fetch
    :rtype: OrderedDict
    :returns: section of a config
    """
    config_filename = join(plugin_dir, PLUGIN_CONFIG_FILENAME)

    with open(config_filename, 'r') as fh:
        config = SafeConfigParser()
        config.readfp(fh)

        if not config.has_section(section):
            return OrderedDict()
        return OrderedDict(config.items(section))


SettingsMeta = namedtuple('SettingsMeta', 'plugin key value default about')


class Command(BaseCommand):
    parser = _parser

    def process(self, argv):
        subcommand = argv.subcommand

        # Handle the actions
        getattr(self, subcommand)(argv)

    def _has_plugin(self, pm, bundle_name, plugin_name):
        """
        Check if a plugin is installed.

        :param string bundle_name: the name of the bundle
        :param string plugin_name: name of the plugin
        :returns: True if it is installed, False if not
        :rtype: bool
        """
        for plugin in pm.plugins:
            if plugin.name == plugin_name and plugin.bundle == bundle_name:
                return True
        return False

    def _settings(self, pm):
        """
        Sorted iteratable of meta information for a plugin setting

        :returns: named tuple of (plugin, setting key, and setting value,
            default value, and about/help message).
        """
        bundles = plugins_by_bundle(pm)

        # Gather up all the settings sorted by plugin and then setting
        sort_bundles = sorted(bundles.items(), key=lambda b: b[0])

        for name, plugins in sort_bundles:
            sort_plugins = sorted(plugins, key=lambda p: p.name)

            for plugin in sort_plugins:
                # Configuration from the plugins.cfg file, not the plugin
                # defaults
                local_config = OrderedDict(plugin.config.items())

                # Get the plugin defaults in case they've changed or vanished
                # from plugins.cfg
                default_config = _get_plugin_config_section(
                    plugin.path, 'settings')

                # About/help messages for the settings
                settings_about = _get_plugin_config_section(
                    plugin.path, 'help')

                # Merge the settings together letting the local settings
                # override the default
                merged_config = OrderedDict()
                merged_config.update(default_config)
                merged_config.update(local_config)

                sort_config = sorted(merged_config.items(), key=lambda s: s[0])

                for key, value in sort_config:
                    yield SettingsMeta(
                        plugin, key, value,
                        default_config.get(key, None),
                        settings_about.get(key, None))

    def list(self, argv):
        """
        List the current settings for all plugins.
        """
        path = argv.path

        with self.out() as out:
            config = get_jigconfig(path)

            pm = PluginManager(config)

            if not pm.plugins:
                out.append(u'No plugins installed.')
                out.extend(NO_PLUGINS_INSTALLED)
                return

            for meta in self._settings(pm):
                out.append(
                    u'{bundle}.{plugin}.{config_key}={config_value}'.format(
                        bundle=meta.plugin.bundle, plugin=meta.plugin.name,
                        config_key=meta.key, config_value=meta.value
                    )
                )

            if not out:
                out.append(u'Installed plugins have no settings.')

            out.extend(CHANGE_PLUGIN_SETTINGS)

    def about(self, argv):
        """
        Provide about/help on each plugin setting.
        """
        path = argv.path

        def wrap(payload):
            indent = '   '
            tw = TextWrapper(
                width=70,
                initial_indent=indent,
                subsequent_indent=indent)
            return u'\n'.join(tw.wrap(payload))

        with self.out() as out:
            config = get_jigconfig(path)

            pm = PluginManager(config)

            if not pm.plugins:
                out.append(u'No plugins installed.')
                out.extend(NO_PLUGINS_INSTALLED)
                return

            for meta in self._settings(pm):
                out.append(u'{bundle}.{plugin}.{config_key}'.format(
                    bundle=meta.plugin.bundle, plugin=meta.plugin.name,
                    config_key=meta.key))
                out.append(u'(default: {0})'.format(meta.default))

                if meta.about:
                    out.append(wrap(meta.about.strip()))

                out.append(u'')

            if not out:
                out.append(u'Installed plugins have no settings.')

    def set(self, argv):
        """
        Change a single setting for an installed plugin.
        """
        path = argv.path
        key = argv.key
        key_parts = key.split('.', 3)
        config_value = argv.value

        with self.out():
            if len(key_parts) != 3:
                # The key is not correct
                raise ConfigKeyInvalid(
                    '{0} is an invalid config key.'.format(key))

            bundle, plugin, config_key = key_parts

            config = get_jigconfig(path)

            pm = PluginManager(config)

            if not self._has_plugin(pm, bundle, plugin):
                raise CommandError('Could not locate plugin {0}.'.format(
                    plugin))

            section_name = 'plugin:{0}:{1}'.format(
                bundle, plugin)

            # Finally change the setting
            pm.config.set(section_name, config_key, config_value)

            set_jigconfig(path, pm.config)



import sys
from stat import S_IXUSR, S_IXGRP, S_IXOTH
from os import stat, chmod, makedirs
from os.path import isfile, isdir, join, realpath, dirname
from shutil import copytree

import git
import gitdb
import async
import smmap

from jig.exc import (
    NotGitRepo, PreCommitExists, JigUserDirectoryError,
    GitTemplatesMissing, GitHomeTemplatesExists, GitConfigError,
    InitTemplateDirAlreadySet)
from jig.conf import JIG_DIR_NAME
from jig.gitutils.scripts import RUN_JIG_SCRIPT, AUTO_JIG_INIT_SCRIPT
from jig.gitutils.checks import is_git_repo

# Dependencies to make jig run
JIG_DIR = realpath(join(dirname(__file__), '..'))
GIT_PYTHON_DIR = realpath(join(dirname(git.__file__), '..'))
GITDB_DIR = realpath(join(dirname(gitdb.__file__), '..'))
ASYNC_DIR = realpath(join(dirname(async.__file__), '..'))
SMMAP_DIR = realpath(join(dirname(smmap.__file__), '..'))


def _git_templates():
    """
    Search and return the location of the shared Git templates directory.

    :rtype: string or None
    """
    search_locations = [
        '/usr/share/git-core/templates',
        '/usr/local/share/git-core/templates',
        '/usr/local/git/share/git-core/templates'
    ]

    for possible_location in search_locations:
        if isdir(possible_location):
            return possible_location

    return None


def _create_pre_commit(destination, template, context):
    """
    Writes a Git pre-commit hook from the template and make it executable.

    :param string destination: the filename of the pre-commit to create
    :param string template: the script template with replaceable vars
        compatible with string.format()
    :param dict context: keys and values to replace in the template
    :raises jig.exc.PreCommitExists: if there is already a Git hook for
        pre-commit present.
    """
    # Is there already a hook?
    if isfile(destination):
        raise PreCommitExists('{0} already exists'.format(destination))

    with open(destination, 'w') as fh:
        fh.write(template.format(**context))

    sinfo = stat(destination)
    mode = sinfo.st_mode | S_IXUSR | S_IXGRP | S_IXOTH

    # Make sure it's executable
    chmod(destination, mode)

    return destination


def hook(gitdir):
    """
    Creates a pre-commit hook that runs Jig in normal mode.

    The hook will be configured to run using the version of Python that was
    used to install jig.

    Returns the full path to the newly created post-commit hook.

    :raises jig.exc.NotGitRepo: if the directory given is not a Git repository.
    """
    if not is_git_repo(gitdir):
        raise NotGitRepo('{0} is not a Git repository.'.format(
            gitdir))

    pc_filename = realpath(join(gitdir, '.git', 'hooks', 'pre-commit'))

    script_kwargs = {
        'python_executable': sys.executable,
        'jig_dir': JIG_DIR,
        'git_python_dir': GIT_PYTHON_DIR,
        'gitdb_dir': GITDB_DIR,
        'async_dir': ASYNC_DIR,
        'smmap_dir': SMMAP_DIR}

    return _create_pre_commit(pc_filename, RUN_JIG_SCRIPT, script_kwargs)


def create_auto_init_templates(user_home_directory):
    """
    Creates a Git templates directory with a Jig auto-init pre-commit hook.

    The templates directory will be created in the user's home directory inside
    a ~/.jig main directory.

    An attempt will be made to find the shared, global Git templates directory
    and copy it. An exception will be thrown if it cannot be found.

    :param string user_home_directory: Full path to the user's home directory
    """
    jig_user_directory = join(user_home_directory, JIG_DIR_NAME)
    jig_git_user_directory = join(jig_user_directory, 'git')

    try:
        map(makedirs, [jig_user_directory, jig_git_user_directory])
    except OSError as ose:
        if ose.errno == 13:
            # Permission denied
            raise JigUserDirectoryError(
                'Cannot create {0} Jig user directory'.format(
                    jig_user_directory
                )
            )
        if ose.errno != 17:
            # Some other kind of OSError
            raise JigUserDirectoryError(unicode(ose))

    # Copy the shared Git templates directory to .jig/git/templates
    git_templates_directory = _git_templates()

    if not git_templates_directory:
        raise GitTemplatesMissing()

    home_templates_directory = join(jig_git_user_directory, 'templates')

    if isdir(home_templates_directory):
        raise GitHomeTemplatesExists(home_templates_directory)

    copytree(git_templates_directory, home_templates_directory)

    pc_filename = realpath(
        join(home_templates_directory, 'hooks', 'pre-commit')
    )

    script_kwargs = {'python_executable': sys.executable}

    _create_pre_commit(
        pc_filename, AUTO_JIG_INIT_SCRIPT, script_kwargs
    )

    return home_templates_directory


def set_templates_directory(templates_directory):
    """
    Sets the template directory in the global Git config.
    """
    command = git.cmd.Git()

    try:
        raw_config = command.config(
            '--global',
            '--list'
        )
    except git.exc.GitCommandError as gce:
        raise GitConfigError(gce)

    config = dict([i.split('=', 1) for i in raw_config.splitlines()])

    if 'init.templatedir' in config:
        raise InitTemplateDirAlreadySet(config['init.templatedir'])

    try:
        command.config(
            '--global',
            '--add',
            'init.templatedir',
            templates_directory
        )
    except git.exc.GitCommandError as gce:
        raise GitConfigError(gce)



# coding=utf-8
from tempfile import mkdtemp

from jig.tests.testcase import CommandTestCase, cd_gitrepo, result_with_hint
from jig.commands.hints import AFTER_INIT
from jig.exc import ForcedExit
from jig.commands import init


class TestInitCommand(CommandTestCase):

    """
    Test the init subcommand.

    """
    command = init.Command

    def test_initialize_repo(self):
        """
        We inititialize a repository with jig.
        """
        self.run_command(self.gitrepodir)

        self.assertResults(
            result_with_hint(
                u'Git repository has been initialized for use with Jig.',
                AFTER_INIT),
            self.output)

    @cd_gitrepo
    def test_initialize_current_directory(self):
        """
        Defaults to the current directory and initializes.
        """
        # Leave the directory argument off, which should make the command
        # use the current working directory.
        self.run_command()

        self.assertResults(
            result_with_hint(
                u'Git repository has been initialized for use with Jig.',
                AFTER_INIT),
            self.output)

    def test_handles_error(self):
        """
        Catches repository that do not exist.
        """
        with self.assertRaises(ForcedExit):
            self.run_command(mkdtemp())

        self.assertIn('is not a Git repository', self.error)
